from typing import List
from abc import *
from thonnycontrib.exceptions import NoFunctionSelectedToTestException
from .ast_parser import L1DocTest, L1TestAstParser
import ast

class TestFinderStrategy(ABC):
    """
    This is an implementation of the Strategy pattern. The strategies defines 
    the way that the ast nodes are extracted. 
    For example, the `FindAllStrategy` strategy will try to find all the ast nodes
    and extract them from the parsed source.
    
    All strategies rely on the AST module of python. The source is first analyzed 
    and all supported nodes are revealed, then each concrete strategy can find 
    its suitable nodes as needed. 
    """
    def __init__(self, filename="", source="", ast_parser:L1TestAstParser=None) -> None:
        self._ast_parser = L1TestAstParser(filename, source) if not ast_parser else ast_parser
    
    # ###### MAIN METHOD ###### #
    def find_l1doctests(self) -> List[L1DocTest]:
        """Invoke this method to get the docstrings depending of the strategy.
        
        This method follows the Template Method pattern. It defines the common 
        algorithm for all concrete strategies. Then, it invokes the `_find_docstring()` 
        abstract method.
        
        Args:
            test_finder (TestFinder): The testFinder fo which the docstring 
            will be retrived.

        Returns:
            List[L1DocTest]: Returns a list of the L1DocTest. Each of L1DocTest 
            represents an AST node.
        """
        ast_nodes = self._ast_parser.parse()
        return self._find_l1doctests(ast_nodes)
    
    @abstractmethod
    def _find_l1doctests(self, ast_nodes: List[ast.AST]=[]) -> List[L1DocTest]:
        """Gets the docstrings from the source detained by the test_finder.

        Args:
            test_finder (TestFinder): An instance of TestFinder. The finder is used 
            to acces it's method to parse the content of the source.
            ast_nodes (List[ast.AST]): The source that is parsed by the TestFinder. 
            Default to empty list.
            
        Returns:
            (List[L1DocTest]): Returns a list of the L1DocTest. Each of L1DocTest 
            represents an AST node.
            
        Raises:
            The overriden methods could throw an exception. For each case, the raised 
            exceptions should be specified and explained into the docstrings.
        """
        pass

    def get_ast_nodes(self):
        return self._ast_nodes 
    
    def set_ast_nodes(self, ast_nodes):
        self._ast_nodes = ast_nodes
    
    def set_filename(self, filename):
        self._ast_parser.set_filename(filename)
    
    def set_source(self, source):
        self._ast_parser.set_source(source)
    
    
class FindAllL1DoctestsStrategy(TestFinderStrategy):
    def _find_l1doctests(self, ast_nodes):
        return self._ast_parser.extract_l1doctests(ast_nodes)


class FindSelectedL1DoctestStrategy(TestFinderStrategy):
    def __init__(self, selected_line) -> None:
        super().__init__()
        self._selected_line = selected_line
    
    def _find_l1doctests(self, ast_nodes):
        """
        Note: The following docstring specifies only the raised excpetion.
        @See the docstring of the superclass to see the full docstring. 
            
        Raises:
            NoFunctionSelectedToTestException: When the selected line doesn' correspond to a 
            function or a class declaration.
        """
        selected_node = self.__find_node(self._selected_line, next(ast_nodes, None))
        return self._ast_parser.extract_l1doctest(selected_node)
    
    def __find_node(self, line, node: ast.AST):
        """ Returns a node by its line number.

        Args:
            line (int): The number of the selected line.
            node (ast.AST): an ast node to be tested. Should be one of the supported types.
            
        Returns:
            ast.AST : Returns an AST node if the node was found for the line number
        
        Raises:
            NoFunctionSelectedToTestException: When no node was found for the line number
        """
        from .ast_parser import SUPPORTED_TYPES
        if node.lineno == line and isinstance(node, SUPPORTED_TYPES) :
            return node
        else: 
            msg = "%s\n\n%s" %  ("No function is selected to test !", 
                                 "The selected line must have a function or a class declaration.")
            raise NoFunctionSelectedToTestException(msg)
    
    def get_selected_line(self):
        return self._selected_line    
    
    def set_selected_line(self, selected_line):
        self._selected_line = selected_line   
        

class L1TestFinder:
    """
        The `TestFinder` relies on the `AST` module to parse the script and extracts 
        the nodes with their docstrings. 
        
        For each node, it invokes the `DoctestParser` to parse its docstring. Then, 
        for each dectected test, it associates an `Example` type that will contains 
        the `source`, `want` and the `line` of the test.
        
        Args:
            - filename(str): the filename of the source code.
            - source(str): the source code to be parsed by the `AST` parser and by the 
                    `DoctestParser`.
            - strategy(TestFinderStrategy): the strategy to use to find the docstrings.
    """
    def __init__(self, filename:str="", source:str="", strategy:TestFinderStrategy=None):      
        self._filename = filename
        self._source = source
        self._strategy = FindAllL1DoctestsStrategy(filename, source) if not strategy else strategy
    
    def find_l1doctests(self) -> List[L1DocTest]: 
        """
        Extract examples from the docstings of all the ast nodes of the source code.
        
        Returns a list of L1DocTest. Each of L1DocTest corresponds to an ast node.
        Each L1DocTest has it's list of `Example`
        
        Raises: 
            NoFunctionSelectedToTestException: when the selected line desn't refer to a 
            function/class signature.
            Error: A compilation error raised by the AST module.
            SpaceMissingAfterPromptException: when a space is missing after the prompt.
        """                
        return self._strategy.find_l1doctests()
        
    def get_filename(self):
        return self._filename
    
    def set_filename(self, filename: str):
        self._filename = filename
        self._strategy.set_filename(filename)
    
    def get_source(self):
        return self._source
    
    def set_source(self, source: str):
        self._source = source
        self._strategy.set_source(source)
        
    def get_strategy(self):
        return self._strategy
    
    def set_strategy(self, strategy: TestFinderStrategy):
        self._strategy = strategy
        self._strategy.set_filename(self._filename)
        self._strategy.set_source(self._source)


from .doc_generator import DocGenerator

# Ceci est une implémentation de création d'un singleton pour L1TestRunner
_doc_generator: DocGenerator = None

def get_doc_generator() -> DocGenerator:
    """
    If there's no `DocGenerator` instance creates one and returns it,
    otherwise returns the current `DocGenerator` instance.
    """
    return DocGenerator() if not _doc_generator else _doc_generator


# ########################################## #
#              Backend exceptions            #
# ########################################## #
class BackendException(BaseException):
    """
        This is the superclass of all the exceptions raised in the
        l1test backend. 
    """
    def get_prefix_message(self) -> str:
        """
            Returns the title to be shown in the error view. 
            The current title is "Cannot run L1Test : ". 

            If it returns `None` so the old title will be kept. 
            Otherwise, it will be changed to the returned value.
        """
        return None

class EmptyValueError(BackendException, ValueError):
    """
    Exception raised when a value is empty. 
    """
    pass

class CompilationError(BackendException): 
    """
    Exception raised when the parsing by the AST parser, `exec()` or by the `compile()` is failed.
    """
    pass

class NoTestFoundException(BackendException):
    """
    Exception raised when an there's no test in the source code. 
    """
    def get_prefix_message(self):
        return ""   
    
class NoFunctionSelectedToTestException(BackendException):
    """
    Exception raised when the selected line doesn't contain a function/class signature.
    """
    pass 

class CannotImportModuleException(BackendException):
    """
    Exception raised when a module cannot be imported by the importlib library.
    """
    pass

class RuntimeException(BackendException):
    """
    Exception raised when a runtime error is raised by the `exec()` method
    """
    pass

class DoctestParserException(BackendException):
    """
    This should be a superclass of all the exceptions raised by the doctest parser
    """
    def __init__(self, filename, lineno, message):
        super().__init__('File %s, line %s\n%s' % (filename, lineno, message))

class SpaceMissingAfterPromptException(DoctestParserException):
    """
    Exception raised when a <space> charatcer is missing after the L1test symbol ($$$) 
    """
    def __init__(self, filename, lineno, invite_sans_blanc):
        super().__init__(filename, lineno, 
                    'A space is missing\n after the prompt : %s' % (invite_sans_blanc)) 
            
            
# ########################################## #
#              Frontend exceptions           #
# ########################################## #
class FrontendException(BaseException):
    """
        This is the superclass of all the exceptions not raised in the
        l1test backend. 
    """
    pass

class NoEditorFoundException(FrontendException):
    """
    exception which includes all exceptions related to the code editor.
    """
    pass
        
class NotSavedFileException(FrontendException):
    """
    Exception raised when the source code is not saved yet on the disk. 
    """
    pass

class EmptyEditorException(FrontendException):
    """
    Exception raised when an editor is still empty. 
    """
    pass

class NoFunctionSelectedToDocumentException(FrontendException):
    """
    Exception raised when no function is selected to document. 
    """
    pass

class DocGeneratorParserException(FrontendException):
    """
    Exception raised when no docstring cannot be generated by the generator.
    This exception is raised when a function signature contains a compilation
    error.
    """
    pass

class CannotSelectSeveralLines(FrontendException):
    """
    Exception raised when several lines are selected. For example, This exception 
    is raised when the user tries to select several lines to run the tests of the selected
    lines. Or, when a user tries to select several lines to generate the docstring.
    """
    pass
